UNIT uTwmSocket;

{$H- shortstring}
{&use32+}
{$i+}

interface
uses
  SysUtils,
  uTwmSockDef,
  uTwmStringS,
{$ifdef win32}
  Windows,
  os2comp,
{$else}
  Dos,
  os2def,
  os2Base,
{$endif}
  uTwmSockFunc;

type
  tSocket = class
    protected
      fDescriptor: tSockDesc;
      fBlocking: boolean;
      fReadBuf: ShortString;
      fReadLogName: string;
      fReadLogHandle: integer;
      fWriteLogName: string;
      fWriteLogHandle: integer;
      procedure SetBlocking(_Blocking: boolean);
      function GetDataAvailable: boolean;
      procedure SetReadLogFile(_ReadLogFile: string);
      procedure SetWriteLogFile(_WriteLogFile: string);
      constructor CreateHot(_Descriptor: tSockDesc; const _Peername: tSockAddr);
    public
      ConnectedTo: tSockAddr;
      BoundTo: TSockAddr;
      constructor Create(_Family, _Type, _Protocol: SmallWord);
      destructor Destroy; override;

      procedure Connect(const _SockAddr: tSockAddr);
      procedure Close;
      function isConnected: boolean;

      procedure Bind(_Family: SmallWord; _InetAddr: ULONG; _Port: SmallWord);
      function isBound: boolean;

      procedure Cancel;

      procedure Listen(_BackLog: ULONG);

      function Accept: tSocket;

      procedure BlockRead(var _Buffer; _BufLen: ULONG; var _BytesRead: ULONG);
      procedure BlockWrite(const _Buffer; _BufLen: ULONG);
      procedure Transact(const _WriteBuffer; _WriteBufLen: ULONG;
                         var _ReadBuffer; _ReadBufLen: ULONG; var _BytesRead: ULONG);
      procedure SendStrLn(_s: AnsiString);
      function RecvStrLn(var _s: AnsiString): boolean;
      function TransactStrLn(_In: AnsiString; var _Out: AnsiString): boolean;

      function BoundToAddrStr: STRING;
      function ConnectedToAddrStr: STRING;

      function BoundToSockStr: STRING;
      function ConnectedToSockStr: STRING;
      function GetLocalHost: string;

      property Blocking: boolean READ fBlocking WRITE SetBlocking;
      property Descriptor: tSockDesc READ fDescriptor;

      PROPERTY DataAvailable: boolean READ GetDataAvailable;
      property ReadLogFile: string read fReadLogName write SetReadLogFile;
      property WriteLogFile: string read fWriteLogName write SetWriteLogFile;

      class function GetHostByAddr(const _InetAddr: tIn_Addr): string;
      class function GetHostByName(const _Hostname: STRING): LongInt;
      class function htons(_Input: SmallInt): SmallInt;
      class function htonl(_Input: LongInt): LongInt;
      class function ntohs(_Input: SmallInt): SmallInt;
      class function ntohl(_Input: LongInt): LongInt;
      class function InetAddrStr(const _InetAddr: tIn_Addr): STRING;
      class function GetServPortByName(_Name, _Proto: STRING): SmallInt;
  end;

implementation

class function tSocket.GetHostByAddr(const _InetAddr: tIn_Addr): string;
  begin
  Result:=SockGetHostNameByAddr(@_InetAddr)
end;

procedure tSocket.SetReadLogFile(_ReadLogFile: string);
  var
    e: eInOutError;
  begin
  if fReadLogHandle>0 then
    FileClose(fReadLogHandle);
  if _ReadLogFile<>'' then begin
    fReadLogHandle:=FileOpen(_ReadLogFile, fmOpenWrite or fmShareDenyWrite);
    if fReadLogHandle=-110 then
      fReadLogHandle:=FileCreate(_ReadLogFile)
    else
      FileSeek(fReadLogHandle, 0, 2);
    if fReadLogHandle<0 then begin
      e:=eInOutError.CreateFmt('Error %d opening file "%s".', [fReadLogHandle, _ReadLogFile]);
      e.ErrorCode:=fReadLogHandle;
      raise e;
    end;
  end else
    fReadlogHandle:=0;
  fReadLogName:=_ReadLogFile;
end;

procedure tSocket.SetWriteLogFile(_WriteLogFile: string);
  var
    e: eInOutError;
  begin
  if fWriteLogHandle>0 then
    FileClose(fWriteLogHandle);
  if _WriteLogFile<>'' then begin
    fWriteLogHandle:=FileOpen(_WriteLogFile, fmOpenWrite or fmShareDenyWrite);
    if fWriteLogHandle=-110 then
      fWriteLogHandle:=FileCreate(_WriteLogFile)
    else
      FileSeek(fWriteLogHandle, 0, 2);
    if fWriteLogHandle<0 then begin
      e:=eInOutError.CreateFmt('Error %d opening file "%s".', [fWriteLogHandle, _WriteLogFile]);
      e.ErrorCode:=fWriteLogHandle;
      raise e;
    end;
  end else
    fWriteLogHandle:=0;
  fWriteLogName:=_WriteLogFile;
end;

function tSocket.GetLocalHost: string;
  var
    sAddr: tSockAddr;
    rc: Longint;
  begin
  rc:=SockGetSockAddr(fDescriptor, sAddr);
  if rc=-1 then
    SockRaiseLastError('SockGetSockName');
  Result:=GetHostByAddr(sAddr.Sin_Addr);
end;

function tSocket.BoundToAddrStr: string;
  begin
  Result := InetAddrStr(BoundTo.Sin_Addr);
end;

function tSocket.ConnectedToAddrStr: STRING;
  begin
  Result := InetAddrStr(ConnectedTo.Sin_Addr);
end;

function tSocket.BoundToSockStr: STRING;
  begin
  Result := BoundToAddrStr + ':' + Long2Dec(htons(BoundTo.Sin_Port));
end;

function tSocket.ConnectedToSockStr: STRING;
  begin
  Result := ConnectedToAddrSTr + ':' + Long2Dec(htons(ConnectedTo.Sin_Port));
end;

constructor tSocket.CreateHot(_Descriptor: tSockDesc; const _Peername: tSockAddr);
  begin
  fDescriptor :=_Descriptor;
  fBlocking := True;
  fReadBuf := '';
  fReadLogName:='';
  fWriteLogName:='';
  FillChar(BoundTo, SizeOf(BoundTo), 0);
  ConnectedTo := _Peername;
end;

function tSocket.Accept: tSocket;
  var
    Res: ApiRet;
    NameLen: ULong;
    PeerName: tSockAddr;
  begin
  NameLen:=SizeOf(Peername);
  Res:=SockAccept(fDescriptor, @Peername, NameLen);
  if Res=-1 then
    SockRaiseLastError('SockAccept');
  Result := tSocket.CreateHot(Res, Peername);
end;

procedure tSocket.Bind(_Family: SmallWord; _InetAddr: ULONG; _Port: SmallWord);
  var
    Res: ApiRet;
    SockAddr: tSockAddr;
  begin
  FillChar(SockAddr, SizeOf(SockAddr), 0);
  SockAddr.Sin_Family := _Family;
  SockAddr.Sin_Addr.IPAddr := _InetAddr;
  SockAddr.Sin_Port := _Port;
  Res := SockBind(fDescriptor, SockAddr);
  if Res=-1 then
    SockRaiseLastError('SockBind');
  if _Port=0 then
    SockGetSockAddr(fDescriptor, BoundTo)
  else
    BoundTo := SockAddr;
end;

function tSocket.isBound: boolean;
  begin
  Result := (BoundTo.Sin_Family <> 0);
end;

procedure tSocket.Cancel;
  begin
  SockCancel(fDescriptor);
end;

procedure tSocket.Listen(_BackLog: ULONG);
  var
    Res: ApiRet;
  begin
  Res := SockListen(fDescriptor, _Backlog);
  if Res=-1 then
    SockRaiseLastError('SockListen');
end;

procedure tSocket.BlockRead(var _Buffer; _BufLen: ULong; var _BytesRead: LongInt);
  var
    p: pChar;
    ReadLen: LongInt;
    WasEmpty: boolean;
  begin
  p:=@_Buffer;
  _BytesRead:=0;
  WasEmpty:=(fReadBuf='');
  if not WasEmpty then begin
    if Length(fReadBuf)>_BufLen then
      _BytesRead:=_BufLen
    else begin
      _BytesRead:=Length(fReadBuf);
      Dec(_BufLen, _BytesRead);
      Inc(p, _BytesRead);
    end;
    Move(fReadBuf[1], _Buffer, _BytesRead);
    Dec(fReadBuf[0], _BytesRead);
  end;
  if WasEmpty or DataAvailable then begin
    ReadLen:=SockRecv(fDescriptor, p, _BufLen, 0);
    if (fReadLogHandle>0) and (ReadLen<>-1) then
      FileWrite(fReadLogHandle, p^, ReadLen);
  end else
    ReadLen:=0;
  if ReadLen=-1 then
    SockRaiseLastError('SockRecv');
  Inc(_BytesRead, ReadLen);
end;

procedure tSocket.BlockWrite(const _Buffer; _BufLen: ULong);
  var
    Res: ApiRet;
  begin
  if fWriteLogHandle>0 then
    FileWrite(fWriteLogHandle, _Buffer, _BufLen);
  Res := SockSend(fDescriptor, @_Buffer, _BufLen, 0);
  if Res=-1 then
    SockRaiseLastError('SockSend');
end;

procedure tSocket.Transact(const _WriteBuffer; _WriteBufLen: ULONG;
                           var _ReadBuffer; _ReadBufLen: ULONG; var _BytesRead: ULONG);
  begin
  BlockWrite(_WriteBuffer, _WriteBufLen);
  BlockRead(_ReadBuffer, _ReadBufLen, _BytesRead);
end;

procedure tSocket.SendStrLn(_s: AnsiString);
  begin
  _s := _s + #13#10;
  UniqueString(_s);
  BlockWrite(_s[1], Length(_s));
end;

function tSocket.RecvStrLn(var _s: AnsiString): boolean;
  var
    BytesRead: INTEGER;
    p: INTEGER;
  begin
  if fReadBuf<>'' then begin
    _s:=fReadBuf;
    fReadBuf:='';
  end else
    _s:='';
  p:=Pos(#13#10, _s);
  if p=0 then begin
    p:=Length(_s);
    SetLength(_s, p+255);
    BlockRead(_s[p+1], 255-p-1, BytesRead);
    SetLength(_s, p+BytesRead);
  end;
  p:=Pos(#13#10, _s);
  Result:=(p>0);
  if Result then begin
    fReadBuf:=TailStr(_s, p+2);
    _s:=LeftStr(_s, p-1);
  end else
    fReadBuf:=_s;
end;

function tSocket.TransactStrLn(_In: AnsiString; var _Out: AnsiString): boolean;
  begin
  SendStrLn(_In);
  while not DataAvailable do
    DosSleep(10);
  Result:=RecvStrLn(_Out);
end;


function tSocket.GetDataAvailable: boolean;
  begin
  Result:=fReadBuf<>'';
  if not Result then
    Result:=(SockSelect(fDescriptor)=1);
end;

procedure tSocket.Close;
  var
    Res: ApiRet;
  begin
  if fDescriptor<>-1 then begin
    Res := SockClose(fDescriptor);
    if Res=-1 then
      SockRaiseLastError('SockClose');
  end;
  fDescriptor := -1;
  FillChar(ConnectedTo, SizeOf(ConnectedTo), 0);
end;

constructor tSocket.Create(_Family, _Type, _Protocol: SmallWord);
  begin
  Inherited Create;
  fBlocking := True;
  fReadBuf := '';
  fReadLogName:='';
  fWriteLogName:='';
  FillChar(ConnectedTo, SizeOf(ConnectedTo), 0);
  FillChar(BoundTo, SizeOf(BoundTo), 0);
  fDescriptor := SockSocket(_Family, _Type, _Protocol);
  if fDescriptor=-1 then
    SockRaiseLastError('SockSocket');
end;

destructor tSocket.Destroy;
  var
    Res: ApiRet;
  begin
  if fReadLogHandle>0 then begin
    FileClose(fReadLogHandle);
    fReadLogHandle:=0;
  end;
  if fWriteLogHandle>0 then begin
    FileClose(fWriteLogHandle);
    fWriteLogHandle:=0;
  end;
  Close;
  Inherited Destroy;
end;

procedure tSocket.Connect(const _SockAddr: tSockAddr);
  var
    Res: ApiRet;
  begin
  Res:=SockConnect(fDescriptor, _SockAddr);
  if Res=-1 then
    SockRaiseLastError('SockConnect');
  ConnectedTo := _SockAddr;
  if BoundTo.Sin_family=0 then
    SockGetSockAddr(fDescriptor, BoundTo)
end;

function tSocket.isConnected: boolean;
  begin
  Result:=ConnectedTo.Sin_Addr.IPAddr <> 0;
end;

procedure tSocket.SetBlocking(_Blocking: boolean);
  begin
  if fBlocking<>_Blocking then begin
    if -1 = SockSetBlockingIo(fDescriptor, _Blocking) then
      SockRaiseLastError('SockSetBlockingIo');
  end;
end;

class function tSocket.GetHostByName(const _Hostname: STRING): LongInt;
  begin
  Result:=SockGetHostAddrByName(_Hostname);
end;

class function tSocket.htons(_Input: SmallInt): SmallInt;
  begin
  Result := SockHtons(_Input);
end;

class function tSocket.htonl(_Input: LongInt): LongInt;
  begin
  Result := SockHtonl(_Input);
end;

class function tSocket.ntohs(_Input: SmallInt): SmallInt;
  begin
  Result:=SockNtohs(_Input);
end;

class function tSocket.ntohl(_Input: LongInt): LongInt;
  begin
  Result:=SockNtohl(_Input);
end;

class function tSocket.InetAddrStr(const _InetAddr: tIn_Addr): STRING;
  begin
  Result := Long2Dec(_InetAddr.ClassA) + '.' +
            Long2Dec(_InetAddr.ClassB) + '.' +
            Long2Dec(_InetAddr.ClassC) + '.' +
            Long2Dec(_InetAddr.ClassD);
end;

class function tSocket.GetServPortByName(_Name, _Proto: STRING): SmallInt;
  begin
  Result := SockGetServPortByName(_Name, _Proto);
end;

end.
